// orders.cpp: 2016-05-17 order(a,p)
// (c) Khashin S.I.  khash2@gmail.com

//  ,   ,     
//     a     p,   2  
//   .

//      unsigned *ord    ,  ,
//   ord[i] = (p-1)/order(a,p).
//    >=2^32,     0   
//  z64.


#ifdef _MSC_VER
#pragma warning(disable : 4996)
#endif

#ifdef __MINGW32__
  #define __USE_MINGW_ANSI_STDIO 1
#endif

#include <stdio.h>

#include "z64.h"			// sieve of Eratosthenes v.3
#include "orders.h"			// order(a,p)

//----------------------------------------------------------------------------
unsigned  orders::a=2;
unsigned  orders::nord=0;
unsigned *orders::ord=NULL;
u64       orders::maxP_=0;
//----------------------------------------------------------------------------
int orders::calc(unsigned a0, u64 maxN)// calculate orders(base)
{
	a=a0;
	if( z64::maxN()<maxN ) return -1;
	if( nord>0 ) delete [] ord;
	nord = (unsigned) z64::Pi(maxN);
	ord = new unsigned[nord];
	ord[0]=0;							// because p(0)=0
	u64 p, q;
	for( unsigned i=1; i<nord; i++ )
	{
		if( i%100000==0 )//|| i>43000 )
			//printf("%7u/%7u K\r", i, nord );
			printf("%7u/%7u K\r", i/1000, nord/1000 );
		p = z64::p(i);
		q = z64::order(a, p);
		if (q == 0) ord[i] = 0;
		else {
			q = (p - 1) / q;
			ord[i] = q <= MAX32 ? (unsigned)q : 0;
		}
	}
	maxP_ = p;
	return 0;
}
//----------------------------------------------------------------------------
u64 orders::ordI(unsigned i)			// order(a, p(i))
{
	if( i<nord ) {
		unsigned q= ord[i];
		return q>0? (z64::p(i)-1)/q : 0;
	}
	return z64::order(a, z64::p(i));
}
u64 orders::ordP(u64 p0)			// order(a, p0)
{
	if (!z64::isPrime(p0)) return 0;
	unsigned ip0 = (unsigned)z64::Pi(p0);
	if( ip0<nord ) {
		unsigned q= ord[ip0];
		if(q>0) return (p0-1)/q;
		return q>0 ? (p0 - 1) / q : 0;
	}
	return z64::order(a, p0);
}
//----------------------------------------------------------------------------
u64 orders::iordP(u64 p0)				// (p0-1)/order(a, p0)
{
	if (!z64::isPrime(p0)) return 0;
	unsigned ip0 = (unsigned)z64::Pi(p0);
	if (ip0<nord) return ord[ip0];
	u64 q = z64::order(a, p0);
	if (q > 0) q = (p0 - 1) / q;
	return q;
}
//----------------------------------------------------------------------------
unsigned orders::iordI(unsigned ip0)			// coorder(p[ip0])
{
	return ip0<nord? ord[ip0]: 0;
}
//----------------------------------------------------------------------------
u64 orders::ordPP(u64 p0, int k)		// order(a, p0^k)
{
	u64 ord = ordP(p0);
	if (ord == 0) return 0;
	u64 pp = p0;
	for (int i = 1; i < k; i++) pp *= p0;	// pp = p0^k
	for( int i=1; i<k; i++ ) {
		if (z64::pow_mod(a, ord, pp) == 1) return ord;
		ord *= p0;
	}
	return ord;
}
//----------------------------------------------------------------------------
u64 orders::ordN(u64 n)				// order(a, n), n - any natural
{
	if (n == 0) return 0;
	if (n == 1) return 1;
	if (n == 2) return (a & 1) == 0 ? 0 : 1;
	// now n>=3
	if (z64::gcd(a, n) != 1) return 0;
	// now n>=3, gcd(a,n)==1
	u64 ord=1;
	while (n > 1) {
		u64 p = z64::primeFactor(n);
		int k = 1;
		n /= p;
		while (n%p == 0) { n /= p; k++; }
		// n   p^k
		u64 ord1 = ordPP(p, k);
		if (ord1 == 0) return 0;
		u64 d = z64::gcd(ord, ord1);
		ord *= ord1 / d;
	}
	return ord;
}
//----------------------------------------------------------------------------
static char fname_buf[256];
char *orders::order_file_name( unsigned a0){
	//sprintf(fname_buf, "c:/a/primes/ord_%04d.dat", a0);
	sprintf(fname_buf, "%sords\\ord_%04d.dat", z64::path, a0);
	return fname_buf;
}
//----------------------------------------------------------------------------
int orders::save(char *fname)	// save orders
{
	FILE *f=fopen(fname, "wb");
	if( !f ) { fclose(f); return 1; }
	unsigned label = 777777777;
	fwrite( &label, sizeof(unsigned), 1, f );
	fwrite( &a,     sizeof(unsigned), 1, f );
	fwrite( &nord,  sizeof(unsigned), 1, f );
	fwrite( ord, 4, nord, f );
	fclose(f);
	return 0;
}
//----------------------------------------------------------------------------
int  orders::load(char *fname, u64 N)	// load orders (base)
{
	//printf(" D01) orders::load(%s, %llu)\n", fname, N);
	if( nord>0) { delete ord; nord=0; }
	FILE *f=fopen(fname, "rb");
	if( !f ) { return 1; }
	unsigned label;
	size_t res;
	res = fread( &label, sizeof(unsigned), 1, f );
	if( res != 1 ) return 2;
	if( label != 777777777 ) return 3;

	res = fread( &a, sizeof(unsigned), 1, f );
	if( res != 1 ) return 2;
	if( a > 100000000 ) return 4;

	res = fread( &nord, sizeof(unsigned), 1, f );
	if( res != 1 ) return 2;
	if( N>0 )
	{
		unsigned iN = (unsigned) z64::Pi(N);
		if( nord>iN ) nord = iN;				//    ,  
	}
	//printf(" D02) nord=%u\n", nord);
	ord = new unsigned[nord];
	res = fread( ord, sizeof(unsigned), nord, f );
	//printf(" D03) res =%u\n", res);
	fclose(f);
	maxP_ = z64::p(nord-1);
	//printf(" D04) nord-1=%u, maxP_=%llu\n", nord-1, maxP_);
	return res == nord? 0 : 5;
}
//----------------------------------------------------------------------------
int  orders::fill(unsigned a0, u64 maxN)	// load seive and orders(2)
{
	char *fname = order_file_name(a0);
	if( load(fname , maxN)==0 ) return 0;
	calc(a0, maxN);
	save(fname);
	return 0;
}
//----------------------------------------------------------------------------
