// pvectors.cpp:    
// 2006-11-06
//
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <vector>
using namespace std;

//#include "point.h"
#include "vUtil.h"
#include "pvectors.h"		//  

//   ,    
void   swapPoints(pnt &p1, pnt &p2);		//  
double angle(double x, double y);			//  (0..2Pi)  (x,y)   X.
double angle(pnt &a, pnt &b, pnt &c);		//  <abc>  0  2*Pi
double kAngle(pnt &a0, pnt &a1);			//   
double Pdistance(pnt &a0, pnt &a1);		//   


////////////////////////////////////////////////////////////////
static double angle(pnt &a, pnt &b, pnt &c)	//  <abc>  0  2*Pi
{
	double x1 = a.x - b.x, y1 = a.y - b.y;
	double x2 = c.x - b.x, y2 = c.y - b.y;
	//      (x1,y1)   (x2,y2)
	// .
	double len1, len2;

	len1 = sqrt(x1*x1 + y1*y1);
	if (len1 == 0) return 0;
	x1 /= len1; y1 /= len1;

	len2 = sqrt(x2*x2 + y2*y2);
	if (len2 == 0) return 0;
	x2 /= len2; y2 /= len2;

	double cosPhi = x1*x2 + y1*y2;
	double sinPhi = x1*y2 - y1*x2;
	return angle(cosPhi, sinPhi);
}

/////////////////////////////////////////////////////////////
//   
//    / =-2
//   *
//    \ =2
static double kAngle(pnt &a0, pnt &a1)
{
	double dx = a1.x - a0.x;
	if (dx == 0) {
		if (a1.y>a0.y) return 1e99; else return -1e99;
	}
	return (a1.y - a0.y) / dx;
}
/////////////////////////////////////////////////////////////
static double Pdistance(pnt &a0, pnt &a1)		//   
{
	int x = a1.x - a0.x, y = a1.y - a0.y;
	return sqrt((double)(x*x + y*y));
}


///////////////////////////////////////////////////////////
// class pvector
//////////////////////////////////////////////////////////////////////
pvector::pvector(int n)
{
	v.resize(n);
}

void pvector::setSize(int n)
{
	v.resize(n);
}

void pvector::add(pvector &w)
{
	//for (int i = 0; i < w.size(); i++) v.push_back(w[i]);
	v.insert(v.end(), w.v.begin(), w.v.end());
}

void pvector::add9(int x, int y)		//  9  (3*3)  
{
	for (int y1 = y - 1; y1 <= y + 1; y1++)
		for (int x1 = x - 1; x1 <= x + 1; x1++)
			add(x1, y1);
}

int pvector::maxX()		//   x-
{
	int mx = 0x80000000;	// min_int
	for (int i = 0; i<this->size(); i++)
		if (this->getX(i)>mx) mx = this->getX(i);
	return mx;
}

int pvector::maxY()		//   y-
{
	int my = 0x80000000;	// min_int
	for (int i = 0; i<this->size(); i++)
		if (this->getY(i)>my) my = this->getY(i);
	return my;
}

int pvector::minX()					//   x-
{
	int mx = 0x7fffffff;
	for (int i = 0; i<this->size(); i++)
		if (this->getX(i)<mx) mx = this->getX(i);
	return mx;
}

int pvector::minY()					//   y-
{
	int my = 0x7fffffff;
	for (int i = 0; i<this->size(); i++)
		if (this->getY(i)<my) my = this->getY(i);
	return my;
}

int	   pvector::indexOf(pnt p)//   p   
{
	for (int i = 0; i<size(); i++)
		if (v[i]==p) return i;
	return -1;
}

//////////////////////////////////////////////////////////////////////////////
// 
//////////////////////////////////////////////////////////////////////////////

enum tpvectorortType	//   
{
	tSort_byX,  		//     X
	tSort_byX_,			//        X
	tSort_byY,			//     Y
	tSort_byY_			//        Y
};

static tpvectorortType pvectorortType = tSort_byX;	//   -    X
static int pvectorSortFlag = 0;						//    
													////////////////////////////////////////////////////////
													//   
int compare_pvector(const void *arg1, const void *arg2)
{
	pnt *p1 = (pnt *)arg1;
	pnt *p2 = (pnt *)arg2;
	switch (pvectorortType)
	{
	case tSort_byX: return p2->x - p1->x;
	case tSort_byX_: return p1->x - p2->x;
	case tSort_byY: return p2->y - p1->y;
	case tSort_byY_: return p1->y - p2->y;
	}
	return 0;
}

void pvector::sortByX()				//        X
{
	if (pvectorSortFlag>0) vError("pvector::sortBy*: pvectorSortFlag>0");
	pvectorSortFlag = 1;
	pvectorortType = tSort_byX;
	qsort(&v[0], this->size(), sizeof(pnt), compare_pvector);
	pvectorSortFlag = 0;
}
void pvector::sortByX_()				//     X
{
	if (pvectorSortFlag>0) vError("pvector::sortBy*: pvectorSortFlag>0");
	pvectorSortFlag = 1;
	pvectorortType = tSort_byX_;
	qsort(&v[0], this->size(), sizeof(pnt), compare_pvector);
	pvectorSortFlag = 0;
}
void pvector::sortByY()				//     Y
{
	if (pvectorSortFlag>0) vError("pvector::sortBy*: pvectorSortFlag>0");
	pvectorSortFlag = 1;
	pvectorortType = tSort_byY;
	qsort(&v[0], this->size(), sizeof(pnt), compare_pvector);
	pvectorSortFlag = 0;
}
void pvector::sortByY_()				//     Y
{
	if (pvectorSortFlag>0) vError("pvector::sortBy*: pvectorSortFlag>0");
	pvectorSortFlag = 1;
	pvectorortType = tSort_byY_;
	qsort(&v[0], this->size(), sizeof(pnt), compare_pvector);
	pvectorSortFlag = 0;
}
//---------------------------------------------------------------
void pvector::createCircle(int x0, int y0, double rad, int mx, int my)	//  .  mx,my>0,        
{
	clear();
	double r2 = rad*rad;
	int r = round(rad);
	bool check = false;
	if (mx>0 || my>0) check = true;

	for (int dy = -r; dy <= r; dy++)
		for (int dx = -r; dx <= r; dx++)
		{
			if (dy*dy + dx*dx>r2) continue;
			int x = x0 + dx, y = y0 + dy;
			if (check && (x<0 || x >= mx || y<0 || y >= my)) continue;
			add(x, y);
		}
	//save("list.txt", "pvector::createCircle");
} // void pvector::createCircle( int x0, int y0, double rad, int mx=0, int my=0)//  .  mx,my>0,        
//---------------------------------------------------------------
void pvector::createBox(int x0, int y0, int dx, int dy) //  
{
	clear();
	for (int y = y0; y<y0 + dy; y++)
		for (int x = x0; x<x0 + dx; x++)
		{
			add(x, y);
		}
	//save("list.txt", "pvector::createBox");
} // void pvector::createCircle( int x0, int y0, double rad, int mx=0, int my=0)	//  .  mx,my>0,        
//---------------------------------------------------------------
//     ,   ,      k:
void  pvector::createCircleK(int x0, int y0, double rad, int K, int mx, int my)	//  .  mx,my>0,        
{
	clear();
	double r2 = rad*rad;
	int r = round(rad);
	bool check = false;
	if (mx>0 || my>0) check = true;

	for (int dy = -r; dy <= r; dy++) {
		int y = y0 + dy;
		if (y%K != 0) continue;
		for (int dx = -r; dx <= r; dx++)
		{
			if (dy*dy + dx*dx > r2) continue;
			int x = x0 + dx;
			if (x%K != 0) continue;
			if (check && (x < 0 || x >= mx || y < 0 || y >= my)) continue;
			add(x, y);
		}
	}
} // void pvector::createCircle( int x0, int y0, double rad, int mx=0, int my=0)//  .  mx,my>0,        
void pvector::createBoxK(int x0, int y0, int dx, int dy, int K)							//  
{
	clear();
	for (int y = y0; y < y0 + dy; y++) {
		if (y%K != 0) continue;
		for (int x = x0; x < x0 + dx; x++)
		{
			if (x%K != 0) continue;
			add(x, y);
		}
	}
	//save("list.txt", "pvector::createBox");
} // void pvector::createCircle( int x0, int y0, double rad, int mx=0, int my=0)	//  .  mx,my>0,        
  //---------------------------------------------------------------
void pvector::save(char *filename, char *msg, int k, char *frm)	//     k      frm
{
	FILE *f = fopen(filename, "w");
	int sz = size();
	fprintf(f, "%4d) %s, #%d\n", k, msg, sz);
	for (int i = 0; i<this->size(); i += k)
	{
		if (k == 1) {
			fprintf(f, "%4d)", i);
			fprintf(f, frm, this->getX(i), this->getY(i));
			fprintf(f, "\n");
		}
		else
		{
			fprintf(f, "%4d(%4d): ", i, i / k);

			for (int ii = 0; ii<k; ii++) {
				if (i + ii >= sz) break;
				fprintf(f, frm, this->getX(i + ii), this->getY(i + ii));
			}
			fprintf(f, "\n");
		}
	}
	fclose(f);
} // void pvector::save( char *filename, char *msg, int k )	//     k   
//---------------------------------------------------------------
void pvector::load(char *filename)				//   
{
	FILE *f = fopen(filename, "r");
	if (!f) vError("pvector::load: file %s not open", filename);
	char stmp[256];
	fgets(stmp, 256, f);
	int i, x, y;
	clear();
	while (EOF != fscanf(f, "%d) (%d,%d)\n", &i, &x, &y))
		add(x, y);
	fclose(f);
} // void pvector::load( char *filename )				//   
//---------------------------------------------------------------
void pvector::convexHull(pvector &p)// this =  (p)
{
	clear();
	int sz = p.size();
	//     
	int min_x = 0x7fffffff;
	int    i, min_i;
	for (i = 0; i<sz; i++) {
		int x = p.getX(i);
		if (x<min_x) { min_x = x; min_i = i; }
	}

	//    .  .
	this->add(p[min_i]);

	//          -
	//      
	double kAngle_min = 2e99;
	int min_diff = 0x7fffffff;
	int min_ii;
	for (i = 0; i<sz; i++) {
		if (i == min_i) continue;
		if (p[i] == get(0)) continue; // ,    - 

		double kA = kAngle(p[min_i], p[i]);
		if (kA<kAngle_min) {
			kAngle_min = kA;
			min_diff = abs(p.getY(min_i) - p.getY(i)) + abs(p.getX(min_i) - p.getX(i));
			min_ii = i;
			continue;
		}
		//    ,    :
		if (kA == kAngle_min) {
			int diff1 = abs(p.getY(min_i) - p.getY(i)) + abs(p.getX(min_i) - p.getX(i));
			if (diff1<min_diff) { min_diff = diff1; min_ii = i; }
		}
	}
	//   .  .
	this->add(p[min_ii]);

	pnt *p0, *p1, *p2;
	double ang, angMax, dist, distMin;
	int max_i;
	do
	{
		p0 = &get(this->size() - 2);
		p1 = &get(this->size() - 1);	//   
		//      p0,p1.    p2 , 
		//  <p0,p1,p2>    .
		//      Pi -   .
		angMax = -1;
		distMin = 1e99;
		for (i = 0; i<sz; i++)
		{
			p2 = &p[i];
			if (*p0==*p2) continue;
			if (*p1==*p2) continue;	//   

			ang = angle(*p2, *p1, *p0);
			if (ang>3.1415927) vError("pvector::convexHull: next angle > Pi");
			if (ang<angMax) continue;
			if (ang == angMax)	//       ,   
			{
				dist = Pdistance(*p1, *p2);		//   
				if (dist>distMin) continue;
			}
			//      
			angMax = ang;
			distMin = Pdistance(*p1, *p2);		//   
			max_i = i;
		}
		p2 = &p[max_i];				//     
		if (*p2 == get(0))break;//       -  
		this->add(*p2);
	} while (1);
} // void pvector::convexHull(pvector &p)// this =  (p)
//---------------------------------------------------------------
double pvector::midPoint(pnt &m)		// m -    p[i],     
{
	int sz = size();
	double sx = 0, sy = 0;
	for (int i = 0; i<sz; i++)
	{
		sx += getX(i);
		sy += getY(i);
	}
	sx /= sz;
	sy /= sz;
	m.x = (int)(sx + 0.5);
	m.y = (int)(sy + 0.5);

	double sum = 0;
	for (int i = 0; i<sz; i++)
	{
		double dx = getX(i) - sx;
		double dy = getY(i) - sy;
		sum += dx*dx + dy*dy;
	}
	return sqrt(sum / sz);
}
double pvector::midPoint(double &x1, double &y1)	// (x1,y1)    p[i],     
{
	int sz = size();
	x1 = y1 = 0;
	for (int i = 0; i<sz; i++)
	{
		x1 += getX(i);
		y1 += getY(i);
	}
	x1 /= sz;
	y1 /= sz;
	double sum = 0;
	for (int i = 0; i<sz; i++)
	{
		double dx = getX(i) - x1;
		double dy = getY(i) - y1;
		sum += dx*dx + dy*dy;
	}
	return sqrt(sum / sz);
}
//////////////////////////////////////////////////////////////////////////////
void pvector::reduceList(pvector &list, int newSize)	// this:=    list ,   ~newSize
{
	int lsz = list.size();
	int k = round(sqrt(lsz / (double)newSize));
	if (k<2) this->v = list.v; //  ,   
	else
	{
		clear();
		for (int i = 0; i<lsz; i++)
		{
			int x = list.getX(i);
			if (x%k != 0) continue;
			int y = list.getY(i);
			if (y%k != 0) continue;
			add(x, y);
		}
		//   ,     
		int ind, newSize9 = newSize - newSize / 16;
		while (size()<newSize9) {
			ind = hRandom(lsz);
			if (this->indexOf(list[ind]) <0)
				this->add(list[ind]);
		}
	}
}
/////////////////////////////////////////////////////////////////////////////////////
//      k (   )
//   mx, my,       (mx,my)
//   ,    , -   ,   
//       ,     k
void pvector::listDivide(pvector &list, int k, int mx, int my) //   listDivide
{
	clear();
	int x, y;
	for (int i = 0; i<list.size(); i++)
	{
		x = list.getX(i);
		y = list.getY(i);
		if (x%k == 0 && y%k == 0) {			//      k,   
			x /= k; y /= k;
			if (mx>0 && x >= mx) continue;
			if (my>0 && y >= my) continue;
			add(x, y);
		}
	}
} // void pvector::listDivide(  pvector &list, int k, int mx, int my) //   listDivide

/*

/////////////////////////////////////////////////////////////////////
void pvector::DeleteDupes()	//     
{
	for( int i=this->size-1; i>0; i-- )
	{
		int ind = this->indexOf( *this->getPoint(i) );
		if( ind<i ) this->DeletePoints( i,i );
	}
}

/////////////////////////////////////////////////////////////////////////
static int iDist( pnt *p, pnt *q )
{
	int dx = p->x - q->x, dy = p->y - q->y;
	return dx*dx + dy*dy;
}
/////////////////////////////////////////////////////////////////////////
void pvector::DeleteAlmostDupes(int r)//    (distance<=r)  
{
	pnt *pi, *pj;
	int r2 = r*r;
	for( int i=this->size-1; i>0; i-- )
	{
		if( i>=this->size-1) continue;
		pi = this->getPoint(i);
		for( int j=0; j<i; j++ )
		{
			pj = this->getPoint(j);
			if( iDist(pi,pj)<=r2 ) 
			{
				pj->x = (pi->x + pj->x)/2;
				pj->y = (pi->y + pj->y)/2;
				this->DeletePoints( i,i );
			}
		}
	}
}






/////////////////////////////////////////////////////////////////////////////
int pvector::mostFarPoint(pnt &a)	//    p    a 
{
	int i, imax;
	double dx, dy, d, dmax=0;
	for( i=0; i<size; i++ )
	{
		dx = getX(i)-a.x;
		dy = getY(i)-a.y;
		d = dx*dx+dy*dy;
		if( d>dmax ) { imax=i; dmax=d; }
	}
	return imax;
}

int pvector::nearestPoint(pnt &a)	//    p     a 
{
	int i, imin;
	double dx, dy, d, dmin=1e99;
	for( i=0; i<size; i++ )
	{
		dx = getX(i)-a.x;
		dy = getY(i)-a.y;
		d = dx*dx+dy*dy;
		if( d<dmin ) { imin=i; dmin=d; }
	}
	return imin;
}

/////////////////////////////////////////////////////////////////////////////////////
//      k (   )
//   mx, my,       (mx,my)
//   ,    , -   ,   
//      { this->listDividQ(list, k, mx, my); return; }
void pvector::listDivide(  pvector &list, int k, int mx, int my) // dst:=src div k
{
	if( k<=1 ) this->CopyFrom(list);
	this->setUniq(0);
	if( list.getSize()/(k*k)>1000 ) { this->listDividQ(list, k, mx, my); return; }
	int x,y;
	this->Empty();
	for( int i=0; i<list.getSize(); i++ )
	{
		x = list.getX(i)/k;
		y = list.getY(i)/k;
		if( mx>0 && x >= mx ) continue;
		if( my>0 && y >= my ) continue;
		if( this->indexOf(pnt(x,y))<0)			//      
			this->AddPoint(x, y);				//  

	}
} // void pvector::listDivide(  pvector &list, int k, int mx, int my) // dst:=src div k
*/
