// 2017-00-10 sq_form.cpp  
#include <math.h>
#include "sq_form.h"	//  

const double DBL_MAX = 1.79769e+308;

//=============================================================================
//  
//=============================================================================
//     9- :
// (-1,-1) ( 0,-1) ( 1,-1)     f[0]   f[1]   f[2]
// (-1, 0) ( 0, 0) ( 1, 0) ->  f[3]   f[4]   f[5]
// (-1, 1) ( 0, 1) ( 1, 1)     f[6]   f[7]   f[8]
//
//       
//
//   f(x,y) = a[0] + a[1]*x + a[2]*y + a[3]*x*x + a[4]*x*y + a[5]*y*y
//
//      
//
//> restart;   -:
//> g : = (x, y)->g0 + g1*x + g2*y + g3*x ^ 2 + g4*x*y + g5*y ^ 2;
//> e00 : = g(-1,-1) - f[0]; e10: = g(0,-1) - f[1]; e20: = g(1,-1) - f[2];
//> e01 : = g(-1, 0) - f[3]; e11: = g(0, 0) - f[4]; e21: = g(1, 0) - f[5];
//> e02 : = g(-1, 1) - f[6]; e12: = g(0, 1) - f[7]; e22: = g(1, 1) - f[8];
//> S : = e00 ^ 2 + e10 ^ 2 + e20 ^ 2
//>      +e01 ^ 2 + e11 ^ 2 + e21 ^ 2
//>      +e02 ^ 2 + e12 ^ 2 + e22 ^ 2;
//> e0: = diff(S, g0); e1: = diff(S, g1); e2: = diff(S, g2); e3: = diff(S, g3); e4: = diff(S, g4); e5: = diff(S, g5);
//> sol: = solve({ e0,e1,e2,e3,e4,e5 }, { g0,g1,g2,g3,g4,g5 });

#define F(x,y) a[0] + a[1]*(x) + a[2]*(y) + a[3]*(x)*(x) + a[4]*(x)*(y) + a[5]*(y)*(y)

double findSqForm(double *f, double *a)	//       a[i]  9-  f[j]
{
	a[0] = (-f[0] + 2 * f[1] - f[2] + 2 * f[3] + 5 * f[4] + 2 * f[5] - f[6] + 2 * f[7] - f[8]) / 9.0;
	a[1] = (-f[0] + f[2] - f[3] + f[5] - f[6] + f[8]) / 6.;
	a[2] = (-f[0] - f[1] - f[2] + f[6] + f[7] + f[8]) / 6.;
	a[3] = (f[0] - 2 * f[1] + f[2] + f[3] - 2 * f[4] + f[5] + f[6] - 2 * f[7] + f[8]) / 6.;
	a[4] = (f[0] - f[2] - f[6] + f[8]) / 4.;
	a[5] = (f[0] + f[1] + f[2] - 2 * f[3] - 2 * f[4] - 2 * f[5] + f[6] + f[7] + f[8]) / 6.;

	double e[9];
	e[0] = F(-1, -1) - f[0];
	e[1] = F(0, -1) - f[1];
	e[2] = F(1, -1) - f[2];

	e[3] = F(-1, 0) - f[3];
	e[4] = F(0, 0) - f[4];
	e[5] = F(1, 0) - f[5];

	e[6] = F(-1, 1) - f[6];
	e[7] = F(0, 1) - f[7];
	e[8] = F(1, 1) - f[8];
	double res = 0;
	for (int i = 0; i < 9; i++) res += e[i] * e[i];
	return sqrt(res)*(1. / 3);
}
//-----------------------------------------------------------------------------
// return:
// 1  (++)  dt>0  a[3]>0
// 2  (+-)  dt<0  
// 3  (--)  dt>0  a[3]<0
// 4  (+0)  dt=0  a[3]>0
// 5  (-0)  dt=0  a[3]<0
// 6  (00)  dt=0  a[3]=0
int    extremSqForm(double *a, double &x0, double &y0, double &z0)	//    a[i]    (x0,y0),    (z0)    
{
	double dt = 4 * a[5] * a[3] - a[4] * a[4];
	if (fabs(dt) < 1e-12) {
		x0 = y0 = DBL_MAX;
		z0 = 0;
		if (a[3]>0) return 4;
		if (a[3]<0) return 5;
		return 6;
	}
	x0 = (a[4] * a[2] - 2 * a[5] * a[1]) / dt;
	y0 = (a[1] * a[4] - 2 * a[3] * a[2]) / dt;
	z0 = a[0] + a[1] * x0 + a[2] * y0 + a[3] * x0*x0 + a[4] * x0*y0 + a[5] * y0*y0;
	if (dt < 0) return 2;
	if (a[3] > 0) return 1;
	return 3;
}
//-----------------------------------------------------------------------------
double vSQ(double *a, double x, double y)		//  6 - . a[i].     (x,y)
{
	return a[0] + a[1] * x + a[2] * y + a[3] * x*x + a[4] * x*y + a[5] * y*y;
}
//-----------------------------------------------------------------------------
bool minSq(double *f, double &x, double &y, double maxOffset)		//    f  9- . (x,y):=  .
{
	double a[6];			// . .
	findSqForm(f, a);
	double z;
	extremSqForm(a, x, y, z);
	bool res = true;
	//if (fabs(x) > maxOffset || fabs(x) > maxOffset) { x = y = maxOffset + 1; }
	if (fabs(x) > maxOffset) {
		double co = maxOffset / fabs(x);
		y *= co;
		x *= co;
		res = false;
	}
	if (fabs(y) > maxOffset) {
		double co = maxOffset / fabs(y);
		y *= co;
		x *= co;
		res = false;
	}
	return res;
}
//-----------------------------------------------------------------------------
